#include "highscore.h"
#include <fcntl.h>
#include <pwd.h>
#include <sys/types.h>
#include <unistd.h>
#include <cstring>
#include <iostream>


std::string GetUserFullName()
{
	const passwd *const user_data (getpwuid(getuid()));
	const char *const res (user_data->pw_gecos);
	if (!res) return user_data->pw_name;
	const char * p;
	for (p = res; *p && *p != ','; ++p);
	return std::string(res, p - res);
}


/*
 *  HighScores
 */

HighScores * HighScores::singleton_ (0);

#define TO_STRING(s) #s
#define MAKE_STRING(s) TO_STRING(s)
#ifndef HIGH_SCORES_FILE
#define HIGH_SCORES_FILE .xgalaga++.scores
#endif
static const char scores_file_name[] = MAKE_STRING(HIGH_SCORES_FILE);


HighScores::HighScores()
{
	// Opening O_RDWR to lockf.
	PosixBuf filebuf (scores_file_name, O_RDONLY | O_CREAT, 0660);
	if (filebuf.FD() < 0) {
		std::cerr << "Cannot open scores file '" << scores_file_name
		          << "'." << std::endl;
		return;		
	}
	if (filebuf.Lock(F_RDLCK) == -1) {
		std::cerr << "Cannot lock scores file '" << scores_file_name
		          << "'." << std::endl;
		return;		
	}
	std::iostream file (&filebuf);
	Load(file);
}


void HighScores::Update(int score, int window_width, int window_height)
{
	PosixBuf filebuf (scores_file_name, O_RDWR | O_CREAT, 0660);
	if (filebuf.FD() < 0) {
		std::cerr << "Cannot open scores file '" << scores_file_name
		          << "'." << std::endl;
		return;		
	}
	if (filebuf.Lock(F_WRLCK) == -1) {
		std::cerr << "Cannot lock scores file '" << scores_file_name
		          << "'." << std::endl;
		return;		
	}
	std::iostream file (&filebuf);
	Load(file);
	high_scores_[std::make_pair(window_width, window_height)].insert(HighScore(score));
	if (!file.eof()) return;
	file.clear();
	if (filebuf.Truncate() != 0) {
		std::cerr << "Cannot reinit scores file '" << scores_file_name
		          << "'." << std::endl;
		return;		
	}
	file.seekg(0);
	Save(file);
}


const std::multiset<HighScore> * HighScores::Get(int window_width,
	                                         int window_height) const
{
	const HSCtr::const_iterator it
		(high_scores_.find(std::make_pair(window_width, window_height)));
	return it == high_scores_.end() ? 0 : &(it->second);
}


void HighScores::Load(std::istream & scores_file)
{
	high_scores_.clear();

	int score;
	int window_width, window_height;
	std::string name;
	while (scores_file >> score >> window_width >> window_height &&
	       std::getline(scores_file, name)) {
		if (!name.empty()) name.erase(0, 1);
		high_scores_[std::make_pair(window_width, window_height)].insert(HighScore(score, name));
	}
}


void HighScores::Save(std::ostream & scores_file) const
{
	for (HSCtr::const_iterator mapit (high_scores_.begin());
	     mapit != high_scores_.end(); ++mapit) {
		for (std::multiset<HighScore>::const_iterator setit (mapit->second.begin());
		     setit != mapit->second.end(); ++setit) {
			scores_file << setit->Value() << ' '
			            << mapit->first.first << ' '
				    << mapit->first.second << ' '
				    << setit->Name() << '\n';
		}
	}
	if (!scores_file.good()) {
		std::cerr << "Error saving scores file '" << scores_file_name
		          << "'." << std::endl;
	}
}
